"""Various constants required by Skyfield."""

from numpy import array

# Definitions.
AU_M = 149597870700             # per IAU 2012 Resolution B2
AU_KM = 149597870.700
ASEC360 = 1296000.0
DAY_S = 86400.0

# Angles.
ASEC2RAD = 4.848136811095359935899141e-6
DEG2RAD = 0.017453292519943296
RAD2DEG = 57.295779513082321
TAU = 6.283185307179586476925287
tau = TAU  # lower case, for symmetry with math.pi

# Physics.
C = 299792458.0

# Earth and its orbit.
ANGVEL = 7.2921150e-5
ERAD = 6378136.6
IERS_2010_INVERSE_EARTH_FLATTENING = 298.25642

PSI_COR = 0.0
EPS_COR = 0.0

# Heliocentric gravitational constant in meters^3 / second^2, from DE-405.
GS = 1.32712440017987e+20

# Time.
T0 = 2451545.0
B1950 = 2433282.4235

C_AUDAY = C * DAY_S / AU_M

# import numpy
# numpy.set_printoptions(formatter={'float': repr})
# from .constants import ASEC2RAD, T0
# from .nutationlib import mean_obliquity
# from .functions import rot_x
# ecliptic_obliquity_radians = mean_obliquity(T0) * ASEC2RAD
# print(repr(rot_x(-ecliptic_obliquity_radians)))

rotation_to_ecliptic = array(((1.0, 0.0, 0.0),
       (0.0, 0.91748214306524178, 0.39777696911260602),
       (0.0, -0.39777696911260602, 0.91748214306524178)))



"""

"""
import os
from datetime import date
from numpy import save
from skyfield import iokit
from skyfield import timelib

_dirname = os.path.dirname(__file__)

cache = iokit.Cache('.')
cache.npy_dirname = _dirname
functions = set([
    timelib.usno_leapseconds,
    ])

def rebuild(remove_old_files=True):
    """Rebuild the data files in this ``skyfield.data`` sub-package.

    You can invoke this routine from the command line with::

        python -m skyfield.data

    """
    if remove_old_files:
        for filename in os.listdir(_dirname):
            if filename.endswith('.npy'):
                os.unlink(os.path.join(_dirname, filename))

    for function in functions:
        return_value = function(cache)
        filename = function.__name__ + '.npy'
        path = os.path.join(_dirname, filename)
        save(path, return_value)

    with open(os.path.join(_dirname, 'rebuild_date.txt'), 'w') as f:
        f.write('%s\n' % date.today())



"""An interface between JPL ephemerides and Skyfield."""

from collections import namedtuple
from numpy import max, min

from jplephem.spk import SPK
from jplephem.names import target_names as _names

from .constants import AU_KM, C_AUDAY, DAY_S
from .errors import DeprecationError
from .functions import length_of
from .positionlib import Astrometric, Barycentric, ICRS
from .timelib import JulianDate, takes_julian_date, calendar_date

Segment = namedtuple('Segment', 'center target compute')
_targets = dict((name, target) for (target, name) in _names.items())


class SpiceKernel(object):
    def __init__(self, filename):
        self.filename = filename
        self.spk = SPK.open(filename)
        self.segments = [Segment(s.center, s.target, _build_compute(s))
                         for s in self.spk.segments]
        self.codes = set(s.center for s in self.segments).union(
                         s.target for s in self.segments)

    def __str__(self):
        segments = self.spk.segments
        lines = ['SPICE kernel file {0!r} has {1} segments'
                 .format(self.filename, len(segments))]
        format_date = '{0}-{1:02}-{2:02}'.format
        start = end = None
        for s in segments:
            if start != s.start_jd or end != s.end_jd:
                start, end = s.start_jd, s.end_jd
                starts = format_date(*calendar_date(int(start)))
                ends = format_date(*calendar_date(int(end)))
                lines.append('    JD {0} - JD {1}  ({2} through {3})'
                             .format(start, end, starts, ends))
            cname = _names.get(s.center, 'unknown')
            tname = _names.get(s.target, 'unknown')
            lines.append('        {0:3} -> {1:<3}  {2} -> {3}'
                         .format(s.center, s.target, cname, tname))
        return '\n'.join(lines)

    def __getitem__(self, name):
        code = self.decode(name)
        return Body(self, code)

    def decode(self, name):
        if isinstance(name, int):
            return name
        name = name.upper()
        code = _targets.get(name)
        if code is None:
            raise KeyError('unknown SPICE target name {0!r}'.format(name))
        if code not in self.codes:
            names = ', '.join(_names[c] for c in self.codes)
            raise KeyError('kernel {0!r} is missing {1!r} -'
                           ' the targets it supports are: {2}'
                           .format(self.filename, name, names))
        return code


def _build_compute(segment):
    """Build a Skyfield `compute` callback for the SPK `segment`."""

    if segment.data_type == 2:
        def compute(jd):
            position, velocity = segment.compute_and_differentiate(jd.tdb)
            return position / AU_KM, velocity / AU_KM

    elif segment.data_type == 3:
        def compute(jd):
            six = segment.compute(jd.tdb)
            return six[:3] / AU_KM, six[3:] * DAY_S / AU_KM

    else:
        raise ValueError('SPK data type {0} not yet supported segment'
                         .format(segment.data_type))
    return compute


class Body(object):
    def __init__(self, ephemeris, code):
        self.ephemeris = ephemeris
        self.segments = ephemeris.segments
        self.code = code

    @takes_julian_date
    def at(self, jd):
        """Compute the Solar System position of this body at a given time."""
        segments = self.segments
        segment_dict = dict((segment.target, segment) for segment in segments)
        chain = list(_center(self.code, segment_dict))[::-1]
        pos, vel = _tally((), chain, jd)
        barycentric = Barycentric(pos, vel, jd)
        barycentric.ephemeris = self.ephemeris
        return barycentric

    def geometry_of(self, body):
        if not isinstance(body, Body):
            code = self.ephemeris.decode(body)
            body = Body(self.ephemeris, code)
        center_chain, target_chain = _connect(self, body)
        return Geometry(self.code, body.code, center_chain, target_chain)

    def _observe_from_bcrs(self, observer):
        return observe(observer, self)

    def topos(self, latitude=None, longitude=None, latitude_degrees=None,
              longitude_degrees=None, elevation_m=0.0, x=0.0, y=0.0):
        assert self.code == 399
        from .toposlib import Topos
        t = Topos(latitude, longitude, latitude_degrees,
                  longitude_degrees, elevation_m, x, y)
        t.ephemeris = self.ephemeris
        t.segments += self.segments
        return t

    def satellite(self, text):
        assert self.code == 399
        from .sgp4lib import EarthSatellite
        lines = text.splitlines()
        return EarthSatellite(lines, self)

    def __call__(self, jd):
        """Deprecated alternative to the new at() method."""
        raise DeprecationError("""use method body.at(jd), not the call body(jd)

If you simply want your old Skyfield script to start working again,
downgrade to Skyfield version 0.4 using a command like:

        pip install skyfield==0.4

Otherwise, you can upgrade your script to modern Skyfield by finding
each place you called a body like a function to generate a position:

        position = body(jd)

Instead, Skyfield now offers a method named at(jd) to makes the
operation easier to read and more symmetrical with other method calls:

        position = body.at(jd)

More documentation can be found at: http://rhodesmill.org/skyfield/""")

def observe(observer, target):
    """Return a light-time corrected astrometric position and velocity.

    Given an `observer` that is a `Barycentric` position somewhere in
    the solar system, compute where in the sky they will see the body
    `target`, by computing the light-time between them and figuring out
    where `target` was back when the light was leaving it that is now
    reaching the eyes or instruments of the `observer`.

    """
    # cposition, cvelocity = _tally([], self.center_chain, jd)
    # tposition, tvelocity = _tally([], self.target_chain, jd)
    jd = observer.jd
    cposition = observer.position.au
    cvelocity = observer.velocity.au_per_d
    tjd = target.at(jd)
    tposition = tjd.position.au
    distance = length_of(tposition - cposition)
    light_time0 = 0.0
    jd_tdb = jd.tdb
    for i in range(10):
        light_time = distance / C_AUDAY
        delta = light_time - light_time0
        if -1e-12 < min(delta) and max(delta) < 1e-12:
            break
        jd2 = JulianDate(tdb=jd_tdb - light_time)
        tjd = target.at(jd2)
        tposition = tjd.position.au
        distance = length_of(tposition - cposition)
        light_time0 = light_time
    else:
        raise ValueError('observe_from() light-travel time'
                         ' failed to converge')
    tvelocity = tjd.velocity.au_per_d
    pos = Astrometric(tposition - cposition, tvelocity - cvelocity, jd)
    pos.light_time = light_time
    pos.observer = observer
    return pos


def _connect(body1, body2):
    """Return ``(sign, segment)`` tuple list leading from body1 to body2."""
    every = body1.segments + body2.segments
    segment_dict = dict((segment.target, segment) for segment in every)
    segments1 = list(_center(body1.code, segment_dict))[::-1]
    segments2 = list(_center(body2.code, segment_dict))[::-1]
    if segments1[0].center != segments2[0].center:
        raise ValueError('cannot trace these bodies back to a common center')
    i = sum(1 for s1, s2 in zip(segments1, segments2) if s1.target == s2.target)
    return segments1[i:], segments2[i:]


def _center(code, segment_dict):
    """Starting with `code`, follow segments from target to center."""
    while code in segment_dict:
        segment = segment_dict[code]
        yield segment
        code = segment.center


class Geometry(object):
    def __init__(self, center, target, center_chain, target_chain):
        self.center = center
        self.target = target
        self.center_chain = center_chain
        self.target_chain = target_chain

    def __str__(self):
        return 'Geometry\n{0}'.format('\n'.join(
            ' {0}'.format(c)
            for c in self.center_chain + self.target_chain))

    @takes_julian_date
    def at(self, jd):
        """Return the geometric Cartesian position and velocity."""
        pos, vel = _tally(self.center_chain, self.target_chain, jd)
        cls = Barycentric if self.center == 0 else ICRS
        return cls(pos, vel, jd)


def _tally(minus_chain, plus_chain, jd):
    position = velocity = 0.0
    for segment in minus_chain:
        p, v = segment.compute(jd)
        position -= p
        velocity -= v
    for segment in plus_chain:
        p, v = segment.compute(jd)
        position += p
        velocity += v
    return position, velocity



"""Basic tests of the Skyfield API module and its contents."""

from assay import assert_raises
from skyfield import api
from skyfield import positionlib

def test_whether_planets_have_radii():
    return # TODO: how will we support this again?
    assert api.mercury.radius.km == 2440.0
    for planet in api.nine_planets:
        assert planet.radius.km > 0.0

def test_sending_jd_that_is_not_a_julian_date():
    earth = api.load('de421.bsp')['earth']
    with assert_raises(ValueError, 'your "jd" argument is not a JulianDate: '):
        earth.at('blah')

def test_apparent_position_class():
    e = api.load('de421.bsp')
    p = e['earth'].at(utc=(2014, 2, 9, 14, 50)).observe(e['mars']).apparent()
    assert isinstance(p, positionlib.Apparent)

def test_astrometric_position_class():
    e = api.load('de421.bsp')
    p = e['earth'].at(utc=(2014, 2, 9, 14, 50)).observe(e['mars'])
    assert isinstance(p, positionlib.Astrometric)

def test_planet_position_class():
    e = api.load('de421.bsp')
    p = e['mars'].at(utc=(2014, 2, 9, 14, 50))
    assert isinstance(p, positionlib.Barycentric)

def test_star_position_class():
    e = api.load('de421.bsp')
    star = api.Star(ra_hours=0, dec_degrees=0)
    p = e['earth'].at(utc=(2014, 2, 9, 15, 1)).observe(star)
    assert isinstance(p, positionlib.Astrometric)

def test_from_altaz_needs_topos():
    p = positionlib.ICRS([0.0, 0.0, 0.0])
    with assert_raises(ValueError, 'the orientation of the horizon'):
        p.from_altaz(alt_degrees=0, az_degrees=0)

def test_from_altaz_parameters():
    e = api.load('de421.bsp')
    usno = e['earth'].topos('38.9215 N', '77.0669 W', elevation_m=92.0)
    p = usno.at(tt=api.T0)
    a = api.Angle(degrees=10.0)
    with assert_raises(ValueError, 'the alt= parameter with an Angle'):
        p.from_altaz(alt='Bad value', alt_degrees=0, az_degrees=0)
    with assert_raises(ValueError, 'the az= parameter with an Angle'):
        p.from_altaz(az='Bad value', alt_degrees=0, az_degrees=0)
    p.from_altaz(alt=a, alt_degrees='bad', az_degrees=0)
    p.from_altaz(az=a, alt_degrees=0, az_degrees='bad')

def test_named_star_throws_valueerror():
    with assert_raises(ValueError, 'No star named foo known to skyfield'):
        api.NamedStar('foo')

def test_named_star_returns_star():
    s = api.NamedStar('Polaris')
    assert isinstance(s, api.Star)



from datetime import date, datetime, timedelta, tzinfo
from numpy import array, einsum, rollaxis, searchsorted, sin, where, zeros_like
from time import strftime
from .constants import T0, DAY_S
from .earthlib import sidereal_time
from .framelib import ICRS_to_J2000 as B
from .nutationlib import compute_nutation, earth_tilt
from .precessionlib import compute_precession

try:
    from pytz import utc
except ImportError:

    class UTC(tzinfo):
        'UTC'
        zero = timedelta(0)
        def utcoffset(self, dt):
            return self.zero
        def tzname(self, dt):
            return 'UTC'
        def dst(self, dt):
            return self.zero

    utc = UTC()

# Much of the following code is adapted from the USNO's "novas.c".

_half_second = 0.5 / DAY_S
_half_millisecond = 0.5e-3 / DAY_S
_half_microsecond = 0.5e-6 / DAY_S
_months = array(['Month zero', 'Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun',
                 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec'])

extra_documentation = """

        This routine takes a date as its argument.  You can either
        provide a `jd=` keyword argument with a `JulianDate` you have
        built yourself, or use one of these keyword arguments::

            # Coordinated Universal Time
            utc=(1973, 12, 29, 23, 59, 48.0)
            utc=datetime(1973, 12, 29, 23, 59, 48.0)

            # International Atomic Time
            tai=2442046.5
            tai=(1973, 12, 29, 23, 59, 48.0)

            # Terrestrial Time
            tt=2442046.5
            tt=(1973, 12, 29, 23, 59, 48.0)

            # Barycentric Dynamical Time
            tdb=2442046.5
            tdb=(1973, 12, 29, 23, 59, 48.0)

"""

def takes_julian_date(function):
    """Wrap `function` so it accepts the standard Julian date arguments.

    A function that takes two arguments, `self` and `jd`, may be wrapped
    with this decorator if it wants to support optional auto-creation of
    its `jd` argument by accepting all of the same keyword arguments
    that the JulianDate constructor itself supports.

    """
    def wrapper(self, jd=None, utc=None, tai=None, tt=None, tdb=None,
                delta_t=0.0, cache=None):
        if jd is None:
            jd = JulianDate(utc, tai, tt, tdb, delta_t, cache)
        elif not isinstance(jd, JulianDate):
            if isinstance(jd, tuple):
                s = _tuple_error
            else:
                s = 'your "jd" argument is not a JulianDate: {0!r}'.format(jd)
            raise ValueError(s)
        return function(self, jd)
    wrapper.__name__ = function.__name__
    synopsis, blank_line, description = function.__doc__.partition('\n\n')
    wrapper.__doc__ = synopsis + extra_documentation + description
    return wrapper

_tuple_error = """was expecting a JulianDate.

Are you trying to pass in a tuple to represent a time? A time scale must
accompany tuples. For example:

    utc=(2014, 1, 18)

Other time scales include tai, tt, and tdb. Please refer to the
```JulianDate``` docstring for more information on these time scales.

"""

def _to_array(value):
    """When `value` is a plain Python sequence, return it as a NumPy array."""
    if not hasattr(value, 'shape') and hasattr(value, '__len__'):
        return array(value)
    else:
        return value

tt_minus_tai = array(32.184 / DAY_S)

class JulianDate(object):
    """A single date and time, or an array, stored as a Julian date.

    You can import this class from ``skyfield.api``.  For the Julian
    date of the current date and time, use the separate function
    ``skyfield.api.now()``.

    Every Julian date object understands four different time scales,
    which can be used during instantiation::

        # Coordinated Universal Time
        JulianDate(utc=(year, month, day, hour, minute, second))
        JulianDate(utc=datetime(year, month, day, hour, minute, second))
        JulianDate(utc=date(year, month, day))

        # International Atomic Time
        JulianDate(tai=2442046.5)  <- Julian day represented by a float
        JulianDate(tai=(year, month, day, hour, minute, second))

        # Terrestrial Time
        JulianDate(tt=2442046.5)  <- Julian day represented by a float
        JulianDate(tt=(year, month, day, hour, minute, second))

        # Barycentric Dynamical Time
        JulianDate(tdb=2442046.5)  <- Julian day represented by a float
        JulianDate(tdb=(year, month, day, hour, minute, second))

    """
    def __init__(self, utc=None, tai=None, tt=None, tdb=None,
                 delta_t=0.0, cache=None):

        self.delta_t = _to_array(delta_t)

        if cache is None:
            from skyfield.data import cache
        self.cache = cache

        if tai is None and utc is not None:
            leap_dates, leap_offsets = cache.run(usno_leapseconds)
            if isinstance(utc, datetime):
                tai = _utc_datetime_to_tai(leap_dates, leap_offsets, utc)
            elif isinstance(utc, date):
                tai = _utc_date_to_tai(leap_dates, leap_offsets, utc)
            elif isinstance(utc, tuple):
                values = [_to_array(value) for value in utc]
                tai = _utc_to_tai(leap_dates, leap_offsets, *values)
            else:
                tai = array([
                    _utc_datetime_to_tai(leap_dates, leap_offsets, dt)
                    for dt in utc])

        if tai is not None:
            if isinstance(tai, tuple):
                tai = julian_date(*tai)
            self.tai = _to_array(tai)
            if tt is None:
                tt = tai + tt_minus_tai

        if tdb is not None:
            if isinstance(tdb, tuple):
                tdb = julian_date(*tdb)
            self.tdb = _to_array(tdb)
            if tt is None:
                tt = tdb - tdb_minus_tt(tdb) / DAY_S

        if tt is None:
            raise ValueError('You must supply either utc, tai, tt, or tdb'
                             ' when building a JulianDate')
        elif isinstance(tt, tuple):
            tt = julian_date(*tt)

        self.tt = _to_array(tt)
        self.shape = getattr(self.tt, 'shape', ())
        self.delta_t = delta_t

    def __repr__(self):
        return '<JulianDate tt={0}>'.format(self.tt)

    def __getitem__(self, index):
        # TODO: also copy cached matrices?
        jd = JulianDate(tt=self.tt[index])
        for name in 'tai', 'tdb', 'ut1', 'delta_t':
            value = getattr(self, name, None)
            if value is not None:
                if getattr(value, 'shape', None):
                    value = value[index]
                setattr(jd, name, value)
        return jd

    def astimezone(self, tz):
        """Return as a Python ``datetime`` in a ``pytz`` provided timezone.

        Convert this Julian date to a ``datetime`` in the timezone `tz`,
        which should be one of the timezones provided by the third-party
        ``pytz`` package.  If this Julian date is an array, then an
        array of datetimes is returned instead of a single value.

        """
        dt, leap_second = self.astimezone_and_leap_second(tz)
        return dt

    def astimezone_and_leap_second(self, tz):
        """Return as a ``datetime`` plus leap second in a ``pytz`` timezone.

        Convert this Julian date to a ``datetime`` and a leap second::

            dt, leap_second = jd.astimezone_and_leap_second(tz)

        The argument `tz` should be a timezone from the third-party
        ``pytz`` package, which must be installed separately.  The date
        and time returned will be for that time zone.

        The leap second value is provided because a Python ``datetime``
        can only number seconds ``0`` through ``59``, but leap seconds
        have a designation of at least ``60``.  The leap second return
        value will normally be ``0``, but will instead be ``1`` if the
        date and time are a UTC leap second.  Add the leap second value
        to the ``second`` field of the ``datetime`` to learn the real
        name of the second.

        If this Julian date is an array, then an array of ``datetime``
        objects and an array of leap second integers is returned,
        instead of a single value each.

        """
        dt, leap_second = self.utc_datetime_and_leap_second()
        normalize = getattr(tz, 'normalize', None)
        if self.shape and normalize is not None:
            dt = array([normalize(d.astimezone(tz)) for d in dt])
        elif self.shape:
            dt = array([d.astimezone(tz) for d in dt])
        elif normalize is not None:
            dt = normalize(dt.astimezone(tz))
        else:
            dt = dt.astimezone(tz)
        return dt, leap_second

    def toordinal(self):
        """Return the proleptic Gregorian ordinal of the TAI date.

        This method makes Skyfield `JulianDate` objects compatible with
        Python `datetime` objects, which also provide a ``toordinal()``
        method.  Thanks to this method, a `JulianDate` can often be used
        directly as a coordinate for a plot.

        """
        return self._utc_float() - 1721424.5

    def utc_datetime(self):
        """Return a Python ``datetime`` for this Julian, expressed as UTC.

        If the third-party ``pytz`` package is available, then its
        ``utc`` timezone will be used as the timezone of the return
        value.  Otherwise, an equivalent Skyfield ``utc`` timezone
        object is used.  If this Julian date is an array, then a
        sequence of datetimes is returned instead of a single value.

        """
        dt, leap_second = self.utc_datetime_and_leap_second()
        return dt

    def utc_datetime_and_leap_second(self):
        """Return a ``datetime`` in UTC, plus a leap second value.

        Convert this Julian date to a ``datetime`` and a leap second::

            dt, leap_second = jd.utc_datetime_and_leap_second()

        If the third-party ``pytz`` package is available, then its
        ``utc`` timezone will be used as the timezone of the return
        value.  Otherwise, Skyfield uses its own ``utc`` timezone.

        The leap second value is provided because a Python ``datetime``
        can only number seconds ``0`` through ``59``, but leap seconds
        have a designation of at least ``60``.  The leap second return
        value will normally be ``0``, but will instead be ``1`` if the
        date and time are a UTC leap second.  Add the leap second value
        to the ``second`` field of the ``datetime`` to learn the real
        name of the second.

        If this Julian date is an array, then an array of ``datetime``
        objects and an array of leap second integers is returned,
        instead of a single value each.

        """
        year, month, day, hour, minute, second = self._utc_tuple(
            _half_millisecond)
        second, fraction = divmod(second, 1.0)
        second = second.astype(int)
        leap_second = second // 60
        second -= leap_second
        milli = (fraction * 1000).astype(int) * 1000
        if self.shape:
            utcs = [utc] * self.shape[0]
            argsets = zip(year, month, day, hour, minute, second, milli, utcs)
            dt = array([datetime(*args) for args in argsets])
        else:
            dt = datetime(year, month, day, hour, minute, second, milli, utc)
        return dt, leap_second

    def utc_iso(self, places=0):
        """Return an ISO 8601 string like ``2014-01-18T01:35:38Z`` in UTC.

        If this Julian date is an array of dates, then a sequence of
        strings is returned instead of a single string.

        """
        if places:
            power_of_ten = 10 ** places
            offset = _half_second / power_of_ten
            year, month, day, hour, minute, second = self._utc_tuple(offset)
            second, fraction = divmod(second, 1.0)
            fraction *= power_of_ten
            format = '%%04d-%%02d-%%02dT%%02d:%%02d:%%02d.%%0%ddZ' % places
            args = (year, month, day, hour, minute, second, fraction)
        else:
            format = '%04d-%02d-%02dT%02d:%02d:%02dZ'
            args = self._utc_tuple(_half_second)

        if self.shape:
            return [format % tup for tup in zip(*args)]
        else:
            return format % args

    def utc_jpl(self):
        """Convert to a string like ``A.D. 2014-Jan-18 01:35:37.5000 UT``.

        Returns a string for this date and time in UTC, in the format
        used by the JPL HORIZONS system.  If this Julian date is an
        array of dates, then a sequence of strings is returned instead
        of a single string.

        """
        offset = _half_second / 1e4
        year, month, day, hour, minute, second = self._utc_tuple(offset)
        second, fraction = divmod(second, 1.0)
        fraction *= 1e4
        bc = year < 1
        year = abs(year - bc)
        era = where(bc, 'B.C.', 'A.D.')
        format = '%s %04d-%s-%02d %02d:%02d:%02d.%04d UT'
        args = (era, year, _months[month], day, hour, minute, second, fraction)

        if self.shape:
            return [format % tup for tup in zip(*args)]
        else:
            return format % args

    def utc_strftime(self, format):
        """Format this UTC time according to a Python date-formatting string.

        This internally calls the Python ``strftime()`` routine from the
        Standard Library ``time()`` module, for which you can find a
        quick reference at ``http://strftime.org/``.  If this Julian
        date is an array of dates, then a sequence of strings is
        returned instead of a single string.

        """
        tup = self._utc_tuple(_half_second)
        year, month, day, hour, minute, second = tup
        second = second.astype(int)
        zero = zeros_like(year)
        tup = (year, month, day, hour, minute, second, zero, zero, zero)
        if self.shape:
            return [strftime(format, item) for item in zip(*tup)]
        else:
            return strftime(format, tup)

    def _utc_tuple(self, offset=0.0):
        """Return UTC as (year, month, day, hour, minute, second.fraction).

        The `offset` is added to the UTC time before it is split into
        its components.  This is useful if the user is going to round
        the result before displaying it.  If the result is going to be
        displayed as seconds, for example, set `offset` to half a second
        and then throw away the fraction; if the result is going to be
        displayed as minutes, set `offset` to thirty seconds and then
        throw away the seconds; and so forth.

        """
        tai = self.tai + offset
        leap_dates, leap_offsets = self.cache.run(usno_leapseconds)
        leap_reverse_dates = leap_dates + leap_offsets / DAY_S
        i = searchsorted(leap_reverse_dates, tai, 'right')
        j = tai - leap_offsets[i] / DAY_S
        whole, fraction = divmod(j + 0.5, 1.0)
        whole = whole.astype(int)
        year, month, day = calendar_date(whole)
        hour, hfrac = divmod(fraction * 24.0, 1.0)
        minute, second = divmod(hfrac * 3600.0, 60.0)
        is_leap_second = j < leap_dates[i-1]
        second += is_leap_second
        return year, month, day, hour.astype(int), minute.astype(int), second

    def _utc_float(self):
        """Return UTC as a floating point Julian date."""
        tai = self.tai
        leap_dates, leap_offsets = self.cache.run(usno_leapseconds)
        leap_reverse_dates = leap_dates + leap_offsets / DAY_S
        i = searchsorted(leap_reverse_dates, tai, 'right')
        return tai - leap_offsets[i] / DAY_S

    def __getattr__(self, name):

        # Cache of several expensive functions of time.

        if name == 'P':
            self.P = P = compute_precession(self.tdb)
            return P

        if name == 'PT':
            self.PT = PT = rollaxis(self.P, 1)
            return PT

        if name == 'N':
            self.N = N = compute_nutation(self)
            return N

        if name == 'NT':
            self.NT = NT = rollaxis(self.N, 1)
            return NT

        if name == 'M':
            self.M = M = einsum('ij...,jk...,kl...->il...', self.N, self.P, B)
            return M

        if name == 'MT':
            self.MT = MT = rollaxis(self.M, 1)
            return MT

        # Conversion between timescales.

        if name == 'tai':
            self.tai = tai = self.tt - tt_minus_tai
            return tai

        if name == 'utc':
            utc = self._utc_tuple()
            utc = array(utc) if self.shape else utc
            self.utc = utc = utc
            return utc

        if name == 'tdb':
            tt = self.tt
            self.tdb = tdb = tt + tdb_minus_tt(tt) / DAY_S
            return tdb

        if name == 'ut1':
            self.ut1 = ut1 = self.tt - self.delta_t / DAY_S
            return ut1

        if name == 'gmst':
            self.gmst = gmst = sidereal_time(self)
            return gmst

        if name == 'gast':
            self.gast = gast = self.gmst + earth_tilt(self)[2] / 3600.0
            return gast

        raise AttributeError('no such attribute %r' % name)

    def __eq__(self, other_jd):
        return self.tt == other_jd.tt


def now():
    """Return the current date and time as a `JulianDate` object.

    For the return value to be correct, your operating system time and
    timezone settings must be set so that the Python Standard Library
    constructor ``datetime.datetime.utcnow()`` returns a correct UTC
    date and time.

    """
    return JulianDate(utc=datetime.utcnow().replace(tzinfo=utc))

def julian_day(year, month=1, day=1):
    """Given a proleptic Gregorian calendar date, return a Julian day int."""
    janfeb = month < 3
    return (day
            + 1461 * (year + 4800 - janfeb) // 4
            + 367 * (month - 2 + janfeb * 12) // 12
            - 3 * ((year + 4900 - janfeb) // 100) // 4
            - 32075)

def julian_date(year, month=1, day=1, hour=0, minute=0, second=0.0):
    """Given a proleptic Gregorian calendar date, return a Julian date float."""
    return julian_day(year, month, day) - 0.5 + (
        second + minute * 60.0 + hour * 3600.0) / DAY_S

def calendar_date(jd_integer):
    """Convert Julian Day `jd_integer` into a Gregorian (year, month, day)."""

    k = jd_integer + 68569
    n = 4 * k // 146097

    k = k - (146097 * n + 3) // 4
    m = 4000 * (k + 1) // 1461001
    k = k - 1461 * m // 4 + 31
    month = 80 * k // 2447
    day = k - 2447 * month // 80
    k = month // 11

    month = month + 2 - 12 * k
    year = 100 * (n - 49) + m + k

    return year, month, day

def tdb_minus_tt(jd_tdb):
    """Computes how far TDB is in advance of TT, given TDB.

    Given that the two time scales never diverge by more than 2ms, TT
    can also be given as the argument to perform the conversion in the
    other direction.

    """
    t = (jd_tdb - T0) / 36525.0

    # USNO Circular 179, eq. 2.6.
    return (0.001657 * sin ( 628.3076 * t + 6.2401)
          + 0.000022 * sin ( 575.3385 * t + 4.2970)
          + 0.000014 * sin (1256.6152 * t + 6.1969)
          + 0.000005 * sin ( 606.9777 * t + 4.0212)
          + 0.000005 * sin (  52.9691 * t + 0.4444)
          + 0.000002 * sin (  21.3299 * t + 5.5431)
          + 0.000010 * t * sin ( 628.3076 * t + 4.2490))

def usno_leapseconds(cache):
    """Download the USNO table of leap seconds as a ``(2, N+1)`` NumPy array.

    The array has two rows ``[leap_dates leap_offsets]``.  The first row
    is used to find where a given date ``jd`` falls in the table::

        index = np.searchsorted(leap_dates, jd, 'right')

    This can return a value from ``0`` to ``N``, allowing the
    corresponding UTC offset to be fetched with::

        offset = leap_offsets[index]

    The offset is the number of seconds that must be added to a UTC time
    to build the corresponding TAI time.

    """
    with cache.open_url('http://maia.usno.navy.mil/ser7/leapsec.dat') as f:
        lines = f.readlines()

    linefields = [line.split() for line in lines]
    dates = [float(fields[4]) for fields in linefields]
    offsets = [float(fields[6]) for fields in linefields]

    dates.insert(0, float('-inf'))
    dates.append(float('inf'))

    offsets.insert(0, offsets[0])
    offsets.insert(1, offsets[0])

    return array([dates, offsets])

def _utc_datetime_to_tai(leap_dates, leap_offsets, dt):
    try:
        utc_datetime = dt.astimezone(utc)
    except ValueError:
        raise ValueError(_naive_complaint)
    tup = utc_datetime.utctimetuple()
    year, month, day, hour, minute, second, wday, yday, dst = tup
    return _utc_to_tai(leap_dates, leap_offsets, year, month, day,
                       hour, minute, second + dt.microsecond / 1000000.00)

def _utc_date_to_tai(leap_dates, leap_offsets, d):
    return _utc_to_tai(leap_dates, leap_offsets, d.year, d.month, d.day)

def _utc_to_tai(leap_dates, leap_offsets, year, month=1, day=1,
                hour=0, minute=0, second=0.0):
    j = julian_day(year, month, day) - 0.5
    i = searchsorted(leap_dates, j, 'right')
    return j + (second + leap_offsets[i]
                + minute * 60.0
                + hour * 3600.0) / DAY_S


_naive_complaint = """cannot interpret a datetime that lacks a timezone

You must either specify that your datetime is in UTC:

    from skyfield.api import utc
    d = datetime(..., tzinfo=utc)  # to build a new datetime
    d = d.replace(tzinfo=utc)      # to fix an existing datetime

Or install the third-party `pytz` library and use any of its timezones:

    from pytz import timezone
    eastern = timezone('US/Eastern')
    d = eastern.localize(datetime(2014, 1, 16, 1, 32, 9))
"""



"""Run a series of timing tests on core Skyfield atoms."""

import gc
import sys
from numpy import array, mean, std, zeros
from skyfield import earthlib, nutationlib, api, starlib

from skyfield.constants import T0
from skyfield.timelib import julian_date, JulianDate
from timeit import default_timer

TA = julian_date(1969, 7, 20, 20., 18.)
TB = julian_date(2012, 12, 21)

D0 = 63.8285
DA = 39.707
DB = 66.8779

earth = api.earth
jupiter = api.jupiter
star = starlib.Star(
    ra_hours=1.59132070233, dec_degrees=8.5958876464,
    ra_mas_per_year=0.0, dec_mas_per_year=0.0,
    parallax_mas=0.0, radial_km_per_s=0.0,
    )


class BM(object):
    def __init__(self, times, bm_fn, t):
        self.name = bm_fn.__name__
        self.times = times
        self.bm_fn = bm_fn
        self.t = t

    def __call__(self):
        self.bm_fn(self.times, self.t)


def run_benchmark(times, fn, *args, **kwargs):
    data = zeros(times)
    for i in xrange(times):
        gc.disable()
        start = default_timer()
        fn(*args, **kwargs)
        end = default_timer()
        gc.enable()

        data[i] = end - start

    avg, stdev, least = mean(data), std(data), min(data)
    suite_name = "{}.{}".format(fn.__module__, fn.__name__)
    factor = 1e6
    print('{} times  {:10.2f} avg  {:10.2f} least  {}'.format(
        times, avg * factor, least * factor, suite_name))


def bm_earth_rotation_angle(times, t):
    run_benchmark(times, earthlib.earth_rotation_angle, t)


def bm_star_observe_from(times, t):
    run_benchmark(times, star.observe_from, earth(t))


def bm_planet_observe_from(times, t):
    run_benchmark(times, jupiter.observe_from, earth(t))


def bm_topo_planet_observe(times, t):
    ggr = earth.topos('75 W', '45 N', elevation_m=0.0)
    run_benchmark(times, ggr(t).observe, jupiter)


def bm_earth_tilt(times, t):
    run_benchmark(times, nutationlib.earth_tilt, t)


def bm_equation_of_the_equinoxes(times, t):
    run_benchmark(times, nutationlib.equation_of_the_equinoxes_complimentary_terms, t)


def bm_fundamental_arguments(times, t):
    run_benchmark(times, nutationlib.fundamental_arguments, t)


def bm_coordinate_to_astrometric(times, t):
    coordinate = star.observe_from(earth(t))
    run_benchmark(times, coordinate.radec)


def bm_coordinate_to_apparent(times, t):
    coordinate = star.observe_from(earth(t))
    run_benchmark(times, coordinate.apparent)


def bm_coordinate_horizontal(times, t):
    ggr = earth.topos('75 W', '45 N', elevation_m=0.0)
    run_benchmark(times, ggr(t).observe(jupiter).apparent().horizontal)


BENCHMARKS = (
    BM(times=100, bm_fn=bm_earth_rotation_angle, t=array([T0, TA, TB])),

    BM(times=100, bm_fn=bm_star_observe_from, t=JulianDate(tt=T0)),
    BM(times=100, bm_fn=bm_star_observe_from, t=JulianDate(tt=TA)),
    BM(times=100, bm_fn=bm_star_observe_from, t=JulianDate(tt=TB)),

    BM(times=100, bm_fn=bm_planet_observe_from, t=JulianDate(tt=T0)),
    BM(times=100, bm_fn=bm_planet_observe_from, t=JulianDate(tt=TA)),
    BM(times=100, bm_fn=bm_planet_observe_from, t=JulianDate(tt=TB)),

    BM(times=100, bm_fn=bm_topo_planet_observe, t=JulianDate(tt=T0)),
    BM(times=100, bm_fn=bm_topo_planet_observe, t=JulianDate(tt=TA)),
    BM(times=100, bm_fn=bm_topo_planet_observe, t=JulianDate(tt=TB)),

    BM(times=100, bm_fn=bm_earth_tilt, t=JulianDate(tt=T0)),
    BM(times=100, bm_fn=bm_earth_tilt, t=JulianDate(tt=TA)),
    BM(times=100, bm_fn=bm_earth_tilt, t=JulianDate(tt=TB)),

    BM(times=100, bm_fn=bm_equation_of_the_equinoxes, t=array([T0, TA, TB])),

    BM(times=100, bm_fn=bm_fundamental_arguments, t=array([T0, TA, TB])),

    BM(times=100, bm_fn=bm_coordinate_to_astrometric, t=JulianDate(tt=T0)),
    BM(times=100, bm_fn=bm_coordinate_to_astrometric, t=JulianDate(tt=TA)),
    BM(times=100, bm_fn=bm_coordinate_to_astrometric, t=JulianDate(tt=TB)),

    BM(times=100, bm_fn=bm_coordinate_to_apparent, t=JulianDate(tt=T0)),
    BM(times=100, bm_fn=bm_coordinate_to_apparent, t=JulianDate(tt=TA)),
    BM(times=100, bm_fn=bm_coordinate_to_apparent, t=JulianDate(tt=TB)),

    BM(times=100, bm_fn=bm_coordinate_horizontal, t=JulianDate(tt=TB)),
)

if __name__ == "__main__":
    patterns = sys.argv[1:]
    for bm in BENCHMARKS:
        if any(pattern not in bm.name for pattern in patterns):
            continue
        bm()



"""Classes representing different kinds of astronomical position."""

from numpy import array, einsum, exp

from .constants import RAD2DEG, TAU, rotation_to_ecliptic
from .functions import from_polar, length_of, to_polar, rot_y, rot_z
from .earthlib import compute_limb_angle, refract, terra
from .relativity import add_aberration, add_deflection
from .timelib import JulianDate, takes_julian_date
from .units import (Distance, Velocity, Angle, _interpret_angle,
                    _interpret_ltude)


class ICRS(object):
    """An x,y,z position whose axes are oriented to the ICRS system.

    The ICRS is a permanent coordinate system that has superseded the
    old series of equinox-based systems like B1900, B1950, and J2000.

    """
    geocentric = True
    altaz_rotation = None

    def __init__(self, position_au, velocity_au_per_d=None, jd=None):
        self.jd = jd
        self.position = Distance(position_au)
        if velocity_au_per_d is None:
            self.velocity = None
        else:
            self.velocity = Velocity(velocity_au_per_d)

    def __repr__(self):
        return '<%s position x,y,z au%s%s>' % (
            self.__class__.__name__,
            '' if (self.velocity is None) else
            ' and velocity xdot,ydot,zdot au/day',
            '' if self.jd is None else ' at date jd',
            )

    def __sub__(self, body):
        """Subtract two ICRS vectors to produce a third."""
        p = self.position.au - body.position.au
        if self.velocity is None or body.velocity is None:
            v = None
        else:
            v = body.velocity.au_per_d - self.velocity.au_per_d
        return ICRS(p, v, self.jd)

    def distance(self):
        """Return the length of this vector.

        >>> v = ICRS([1.0, 1.0, 0.0])
        >>> print(v.distance())
        1.41421 au

        """
        return Distance(length_of(self.position.au))

    def radec(self, epoch=None):
        """Return this position as a tuple (RA, declination, distance).

        >>> ra, dec, distance = ICRS([1.0, 1.0, 1.0]).radec()
        >>> ra
        <Angle 03h 00m 00.00s>
        >>> dec
        <Angle +35deg 15' 51.8">
        >>> distance
        <Distance 1.73205 au>

        """
        position_au = self.position.au
        if epoch is not None:
            if isinstance(epoch, JulianDate):
                pass
            elif isinstance(epoch, float):
                epoch = JulianDate(tt=epoch)
            elif epoch == 'date':
                epoch = self.jd
            else:
                raise ValueError('the epoch= must be a Julian date,'
                                 ' a floating point Terrestrial Time (TT),'
                                 ' or the string "date" for epoch-of-date')
            position_au = einsum('ij...,j...->i...', epoch.M, position_au)
        r_au, dec, ra = to_polar(position_au)
        return (Angle(radians=ra, preference='hours'),
                Angle(radians=dec, signed=True),
                Distance(r_au))

    def ecliptic_position(self):
        """Return an x,y,z position relative to the ecliptic plane."""
        vector = rotation_to_ecliptic.dot(self.position.au)
        return Distance(vector)

    def ecliptic_latlon(self):
        """Return ecliptic latitude, longitude, and distance."""
        vector = rotation_to_ecliptic.dot(self.position.au)
        d, lat, lon = to_polar(vector)
        return (Angle(radians=lat, signed=True),
                Angle(radians=lon),
                Distance(au=d))

    def from_altaz(self, alt=None, az=None, alt_degrees=None, az_degrees=None):
        """Generate an Apparent position from an altitude and azimuth.

        The altitude and azimuth can each be specified by an `Angle`
        object or else explicitly with a degree measurement:

        alt=Angle(...), az=Angle(...)
        alt_degrees=23.2289, az_degrees=142.1161
        alt_degrees=(23, 13, 44.1), az_degrees=(142, 6, 58.1)

        """
        R = self.altaz_rotation
        if R is None:
            raise ValueError('only a position generated by a topos() call'
                             ' knows the orientation of the horizon'
                             ' and can understand altitude and azimuth')
        alt = _interpret_angle('alt', alt, alt_degrees)
        az = _interpret_angle('az', az, az_degrees)
        r = 0.1  # close enough to make gravitational refraction irrelevant
        p = from_polar(r, alt, az)
        p = einsum('ji...,j...->i...', R, p)
        return Apparent(p)


class Topos(object):
    """An object representing a specific location on the Earth's surface."""

    def __init__(self, latitude=None, longitude=None, latitude_degrees=None,
                 longitude_degrees=None, elevation_m=0.0):

        if latitude_degrees is not None:
            latitude = Angle(degrees=latitude_degrees)
        elif isinstance(latitude, str):
            latitude = _interpret_ltude(latitude, 'latitude', 'N', 'S')
        elif not isinstance(latitude, Angle):
            raise TypeError('please provide either latitude_degrees=<float>'
                            ' or latitude=<skyfield.units.Angle object>'
                            ' with north being positive')

        if longitude_degrees is not None:
            longitude = Angle(degrees=longitude_degrees)
        elif isinstance(longitude, str):
            longitude = _interpret_ltude(longitude, 'longitude', 'E', 'W')
        elif not isinstance(longitude, Angle):
            raise TypeError('please provide either longitude_degrees=<float>'
                            ' or longitude=<skyfield.units.Angle object>'
                            ' with east being positive')

        self.latitude = latitude
        self.longitude = longitude
        self.elevation = Distance(m=elevation_m)

        self.R_lat = rot_y(latitude.radians)[::-1]

    @takes_julian_date
    def __call__(self, jd):
        """Compute where this Earth location was in space on a given date."""
        e = self.ephemeris.earth(jd)
        tpos_au, tvel_au_per_d = self._position_and_velocity(jd)
        t = Barycentric(e.position.au + tpos_au,
                        e.velocity.au_per_d + tvel_au_per_d,
                        jd)
        t.geocentric = False  # test, then get rid of this attribute
        t.rGCRS = tpos_au
        t.vGCRS = tvel_au_per_d
        t.topos = self
        t.ephemeris = self.ephemeris
        t.altaz_rotation = self._altaz_rotation(jd)
        return t

    @takes_julian_date
    def gcrs(self, jd):
        """Compute where this location was in the GCRS on a given date."""
        tpos_au, tvel_au_per_d = self._position_and_velocity(jd)
        t = Geocentric(tpos_au, tvel_au_per_d, jd)
        t.topos = self
        t.ephemeris = self.ephemeris
        t.altaz_rotation = self._altaz_rotation(jd)
        return t

    def _position_and_velocity(self, jd):
        """Return the GCRS position, velocity of this Topos at `jd`."""
        pos, vel = terra(self.latitude.radians, self.longitude.radians,
                         self.elevation.au, jd.gast)
        pos = einsum('ij...,j...->i...', jd.MT, pos)
        vel = einsum('ij...,j...->i...', jd.MT, vel)
        return pos, vel

    def _altaz_rotation(self, jd):
        """Compute the rotation from the ICRS into the alt-az system."""
        R_lon = rot_z(- self.longitude.radians - jd.gast * TAU / 24.0)
        return einsum('ij...,jk...,kl...->il...', self.R_lat, R_lon, jd.M)


class Barycentric(ICRS):
    """BCRS: an ICRS x,y,z position measured from the Solar System barycenter.

    """
    def observe(self, body):
        """Return the astrometric position of `body` viewed from this position.

        """
        astrometric = body._observe_from_bcrs(self)
        return astrometric


class Astrometric(ICRS):
    """An astrometric position as an x,y,z vector in the ICRS.

    The *astrometric position* of a body is its position relative to an
    observer, adjusted for light-time delay: the position of the body
    back when it emitted (or reflected) the light that is now reaching
    the observer's eyes or telescope.  This is always a difference
    between two BCRS vectors.

    """
    def apparent(self):
        """Return the apparent position where this will appear in the sky.

        This method determines how relativity affects an image, and
        returns the :class:`~skyfield.positionlib.Apparent` position
        where the body will actually appear in the sky.  The effects
        modeled are the deflection that the image will experience if its
        light passes close to large masses in the Solar System, and the
        aberration caused by the observer's own velocity.

        These transforms convert the position from the BCRS reference
        frame of the Solar System barycenter and to the reference frame
        of the observer.  In the specific case of an Earth observer, the
        output reference frame is the GCRS.

        """
        jd = self.jd
        position_au = self.position.au.copy()
        observer = self.observer

        if observer.geocentric:
            include_earth_deflection = array((False,))
        else:
            limb_angle, nadir_angle = compute_limb_angle(
                position_au, observer.position.au)
            include_earth_deflection = limb_angle >= 0.8

        add_deflection(position_au, observer.position.au, observer.ephemeris,
                       jd.tdb, include_earth_deflection)
        add_aberration(position_au, observer.velocity.au_per_d, self.light_time)

        a = Apparent(position_au, jd=jd)
        a.observer = self.observer
        return a


class Apparent(ICRS):
    """An apparent position as an x,y,z vector in the GCRS.

    The *apparent position* of a body is its position relative to an
    observer, adjusted not only for the light-time delay between the
    body and an observer (which was already accounted for in the
    object's astrometric position), but also adjusted for deflection
    (its light rays bending as they pass large masses like the Sun or
    Jupiter) and aberration (light slanting because of the observer's
    motion through space).

    Included in aberration is the relativistic transformation that takes
    the position out of the BCRS centered on the solar system barycenter
    and into the GCRS centered on the Earth.

    If the observer was a planet or satellite with its own orbit around
    the Sun, then this apparent position is not really a GCRS position,
    but belongs to a GCRS-like system centered on that observer instead.

    """
    def altaz(self, temperature_C=None, pressure_mbar='standard'):
        """Return the position as a tuple ``(alt, az, distance)``.

        `alt` - Altitude in degrees above the horizon.
        `az` - Azimuth angle east around the horizon from due-north.
        `distance` - Distance to the object.

        """
        try:
            topos = self.observer.topos
            R = self.observer.altaz_rotation
        except AttributeError:
            raise ValueError('to compute an apparent position, you must'
                             ' observe from a specific Earth location that'
                             ' you specify using a Topos instance')

        # TODO: wobble

        position_au = einsum('ij...,j...->i...', R, self.position.au)
        r_au, alt, az = to_polar(position_au)

        if temperature_C is None:
            alt = Angle(radians=alt)
        else:
            if temperature_C == 'standard':
                temperature_C = 10.0
            if pressure_mbar == 'standard':
                pressure_mbar = 1010.0 * exp(-topos.elevation.m / 9.1e3)
            alt = refract(alt * RAD2DEG, temperature_C, pressure_mbar)
            alt = Angle(degrees=alt)

        return alt, Angle(radians=az), Distance(r_au)


class Geocentric(ICRS):
    """A position referred to the GCRS as measured from the geocenter."""

    def observe(self, other):
        gcrs_method = getattr(other, 'gcrs')
        if gcrs_method is None:
            raise ValueError('currently a Geocentric location can only'
                             ' observe an object that can generate a'
                             ' GCRS position through a .gcrs() method')
        g = gcrs_method(self.jd)
        # TODO: light-travel-time backdating, if distant enough?
        p = g.position.au - self.position.au
        v = g.velocity.au_per_d - self.velocity.au_per_d
        a = Apparent(p, v, self.jd)
        a.observer = self
        return a


def ITRF_to_GCRS(jd, rITRF):  # todo: velocity

    # Todo: wobble

    spin = rot_z(jd.gast * TAU / 24.0)
    position = einsum('ij...,j...->i...', spin, array(rITRF))
    return einsum('ij...,j...->i...', jd.MT, position)



"""Helpers for making Skyfield tests stable."""

from functools import partial

import skyfield.api

def setup(utc=(2015, 10, 11, 10)):
    def now():
        """Return a constant "now"."""
        return skyfield.api.JulianDate(utc=utc)
    skyfield.api.now = now
    skyfield.api.load = partial(skyfield.api.load, verbose=False)

def teardown():
    skyfield.api.now = skyfield.timelib.now
    skyfield.api.load = skyfield.iokit.load



"""Raw transforms between coordinate frames, as NumPy matrices."""

from numpy import array
from .constants import ASEC2RAD

def build_matrix():
    # 'xi0', 'eta0', and 'da0' are ICRS frame biases in arcseconds taken
    # from IERS (2003) Conventions, Chapter 5.

    xi0  = -0.0166170 * ASEC2RAD
    eta0 = -0.0068192 * ASEC2RAD
    da0  = -0.01460   * ASEC2RAD

    # Compute elements of rotation matrix.

    yx = -da0
    zx =  xi0
    xy =  da0
    zy =  eta0
    xz = -xi0
    yz = -eta0

    # Include second-order corrections to diagonal elements.

    xx = 1.0 - 0.5 * (yx * yx + zx * zx)
    yy = 1.0 - 0.5 * (yx * yx + zy * zy)
    zz = 1.0 - 0.5 * (zy * zy + zx * zx)

    return array(((xx, xy, xz), (yx, yy, yz), (zx, zy, zz)))

ICRS_to_J2000 = build_matrix()
del build_matrix



"""A work in progress, that does not yet agree with JPL Horizons.

See https://github.com/brandon-rhodes/python-skyfield/issues/11

"""
from math import sin, cos
import math
from . import constants

# class to represent a point in the IC reference frame
class ICRCoordinates:
    def __init__(self, x, y, z):
        self.x = x
        self.y = y
        self.z = z

    def equalTo(self, other):
        # TODO: override ==, and add epsilons here
        return (self.x == other.x) and (self.y == other.y) and (self.z == other.z)

    def __repr__(self):
        return '(%s, %s, %s)' % (self.x, self.y, self.z)

def semimajorAxisToOrbitalPeriod(axis):
    return (axis ** 3) ** 0.5

def orbitalPeriodToSemimajorAxis(period):
    return (period ** 2) ** (1.0 / 3.0)

def convergeEccentricAnomaly(mean_anomaly, eccentricity, precision):
    # calculate the delta
    delta = 10 ** -precision

    # normalize the mean anomaly
    m = mean_anomaly % constants.TAU

    # set up the first guess
    eccentric_anomaly = constants.TAU
    if eccentricity < 0.8:
        eccentric_anomaly = m

    # do the initial test
    test = eccentric_anomaly - eccentricity * sin(m) - m

    # while we're not happy with the result, and we haven't been dawdling too long
    max_iterations = 30
    count = 0
    while ((math.fabs(test) > delta) and (count < max_iterations)):
        # calculate the next guess for an eccentric anomaly
        eccentric_anomaly = (
            eccentric_anomaly - test /
            (1.0 - eccentricity * cos(eccentric_anomaly))
        )

        # try it
        test = eccentric_anomaly - eccentricity * sin(eccentric_anomaly) - m

        # count the runs, so we don't go forever
        count += 1

    # convert to degrees
    return eccentric_anomaly

def calculateMeanAnomaly(L, wb):
    return L - wb

class KeplerianOrbit:
    def __init__(
            self,
            semimajor_axis,
            eccentricity,
            inclination,
            longitude_ascending,
            argument_perihelion,
            mean_anomaly,
            epoch
        ):
        self.semimajor_axis = semimajor_axis
        self.eccentricity = eccentricity
        self.inclination = inclination
        self.longitude_ascending = longitude_ascending
        self.argument_perihelion = argument_perihelion
        self.mean_anomaly = mean_anomaly
        self.epoch = epoch

    def getECLCoordinatesOnJulianDate(self, date):
         # localize the orbital parameters
        a = self.semimajor_axis
        e = self.eccentricity
        I = self.inclination
        Om = self.longitude_ascending
        #n = 0.230605479
        n = 0.230652907

        w = self.argument_perihelion

        M = self.mean_anomaly
        d = date.tdb - self.epoch.tdb

        Om = Om / 360.0 * constants.TAU
        w = w / 360.0 * constants.TAU
        I = I / 360.0 * constants.TAU
        M = M / 360.0 * constants.TAU
        n = n / 360.0 * constants.TAU

        M += d * n

        # calculate the mean anomaly in rads
        E = convergeEccentricAnomaly(M, e, 30)

        # calculate the initial primes
        x_prime = a * (cos(E) - e)
        y_prime = a * (1 - e ** 2.0) ** (0.5) * sin(E)

        """
        http://ssd.jpl.nasa.gov/txt/aprx_pos_planets.pdf
        x_ecl = cos(w)cos(Om)-sin(w)sin(Om)cos(I) * x_prime +
            (-sin(w)cos(Om) - cos(w)sin(Om)cos(I)) * y_prime
        y_ecl = cos(w)sin(Om)-sin(w)cos(Om)cos(I) * x_prime +
            (-sin(w)cos(Om) - cos(w)sin(Om)cos(I)) * y_prime
        z_ecl = (sin(w)sin(I)) * x_prime +
            (cos(w)sin(I)) * y_prime
        """

        # calculate the ecliptic coordinates
        x_ecl = ((cos(w) * cos(Om) - sin(w) * sin(Om) * cos(I)) * x_prime +
                (-1 * sin(w) * cos(Om) - cos(w) * sin(Om) * cos(I)) * y_prime)
        y_ecl = ((cos(w) * sin(Om) + sin(w) * cos(Om) * cos(I)) * x_prime +
                (-1 * sin(w) * sin(Om) + cos(w) * cos(Om) * cos(I)) * y_prime)
        z_ecl = ((sin(w) * sin(I)) * x_prime + (cos(w) * sin(I)) * y_prime)

        return ICRCoordinates(x_ecl, y_ecl, z_ecl)

    def getICRSCoordinatesOnJulianDate(self, date):
        # J2000 obliquity
        e = 23.43928 * math.pi / 180.0

        # get the ecliptic coords
        ecliptic = self.getECLCoordinatesonJulianDate(date);

        # calculate the equatorial (ICRS) coordinates
        x_eq = ecliptic.x;
        y_eq = cos(e) * ecliptic.y - sin(e) * ecliptic.z
        z_eq = sin(e) * ecliptic.y + cos(e) * ecliptic.z




from __future__ import print_function
import os
import numpy as np
import sys
from datetime import datetime, timedelta
from time import time

from .jpllib import SpiceKernel

try:
    from fcntl import LOCK_EX, LOCK_UN, lockf
except:
    lockf = None

try:
    from urllib.request import urlopen, url2pathname
except:
    from urllib2 import urlopen, url2pathname

_missing = object()


def load(filename, autodownload=True, verbose=True):
    """Load the given file, possibly downloading it if it is not present."""
    if filename.endswith('.bsp'):
        url = url_for(filename)
        cls = SpiceKernel
    else:
        raise ValueError('Skyfield does not recognize that file extension')
    if not os.path.exists(filename):
        if not autodownload:
            raise IOError('you specified autodownload=False but {!r} cannot'
                          ' be found in the current directory'
                          .format(filename))
        url = url_for(filename)
        download(url, verbose)
    return cls(filename)


def url_for(filename):
    """Given a recognized filename, return its URL."""
    if filename.endswith('.bsp'):
        if filename.startswith('de'):
            return 'ftp://ssd.jpl.nasa.gov/pub/eph/planets/bsp/' + filename
        elif filename.startswith('jup'):
            return ('http://naif.jpl.nasa.gov/pub/naif/generic_kernels'
                    '/spk/satellites/' + filename)
        raise ValueError('Skyfield does not know where to download {!r} from'
                         .format(filename))


def download(url, verbose=True, blocksize=128*1024):
    """Download a file from a URL, possibly displaying a progress bar."""
    filename = os.path.basename(url2pathname(url))
    if os.path.exists(filename):
        return filename
    tempname = filename + '.download'
    try:
        connection = urlopen(url)
    except Exception as e:
        raise IOError('cannot fetch file {0!r} from URL {1} because {2}'
                      .format(filename, url, e))
    content_length = int(connection.headers.get('content-length', -1))
    if verbose:
        bar = ProgressBar(filename)
    with open(tempname, 'ab') as w:
        try:
            if lockf is not None:
                fd = w.fileno()
                lockf(fd, LOCK_EX)
                if os.fstat(fd).st_size:
                    return  # someone else wrote the file contents
            length = 0
            while True:
                data = connection.read(blocksize)
                if not data:
                    break
                w.write(data)
                length += len(data)
                if verbose:
                    bar.report(length, content_length)
            w.flush()
            os.rename(tempname, filename)
        except KeyboardInterrupt:# Exception as e:
            raise IOError('error getting {0} - {1}'.format(url, e))
        finally:
            if lockf is not None:
                lockf(fd, LOCK_UN)
    return filename


class ProgressBar(object):
    def __init__(self, filename):
        self.filename = filename
        self.t0 = 0

    def report(self, bytes_so_far, bytes_total):
        if bytes_total < 0:
            return
        percent = 100 * bytes_so_far // bytes_total
        if (percent != 100) and (time() - self.t0 < 0.5):
            return
        self.t0 = time()
        bar = '#' * (percent // 3)
        print('\r[{0:33}] {1:3}% {2}'.format(bar, percent, self.filename),
              end='\n' if (percent == 100) else '')
        sys.stdout.flush()


class Cache(object):
    """Early experiment in being sensitive to file dates and age."""
    def __init__(self, cache_path, days_old=0):
        self.cache_path = cache_path
        self.days_old = days_old
        self.ram_cache = {}
        self.npy_dirname = None

    def open_url(self, url, days_old=None):
        filename = url[url.rindex('/') + 1:]
        path = os.path.join(self.cache_path, filename)
        if days_old is None:
            days_old = self.days_old
        download_file(url, path, days_old)
        return open(path, 'rb')

    def run(self, function):
        """Return the result of running `function(this_cache)` one time only.

        If this cache has already been asked to run `function`, then the
        return value of its first run is returned without re-running it.

        """
        result = self.ram_cache.get(function, _missing)
        if result is not _missing:
            return result

        if self.npy_dirname:
            path = os.path.join(self.npy_dirname, function.__name__ + '.npy')
            if os.path.exists(path):
                # TODO: check whether data is recent enough
                result = np.load(path)
                self.ram_cache[function] = result
                return result

        result = function(self)
        self.ram_cache[function] = result
        return result


def download_file(url, filename, days_old=0):
    """Early experiment in what download logic might look like."""
    if os.path.exists(filename):
        if not is_days_old(filename, days_old):
            return

    response = urlopen(url)
    f = open(filename, 'wb')
    while True:
        block = response.read(4096)
        if not block:
            break
        f.write(block)

    f.close()

def is_days_old(filename, days_old):
    """Early experiment in being sensitive to file dates."""
    min_old = datetime.now()-timedelta(days=days_old)
    modified = datetime.fromtimestamp(os.path.getmtime(filename))
    return modified < min_old



import numpy as np
from assay import assert_raises
from pytz import timezone
from skyfield.constants import DAY_S
from skyfield.timelib import JulianDate, takes_julian_date, utc
from datetime import datetime

one_second = 1.0 / DAY_S
epsilon = one_second * 42.0e-6  # 20.1e-6 is theoretical best precision

time_parameter = ['tai', 'tt', 'tdb']
time_value = [(1973, 1, 18, 1, 35, 37.5), 2441700.56640625]

def test_tuple_error_raised():
    def fake_function():
        """ The decorator requires a docstring. """
    decorated_function = takes_julian_date(fake_function)
    tuple_argument = (2014, 1, 23)
    self = None
    with assert_raises(ValueError) as info:
        decorated_function(self, tuple_argument)
    assert 'Are you trying to pass in a tuple' in str(info.exception)

def test_JulianDate_init(time_parameter, time_value):
    kw = {time_parameter: time_value}
    jd = JulianDate(**kw)
    assert getattr(jd, time_parameter) == 2441700.56640625

def test_building_JulianDate_from_utc_tuple_with_array_inside():
    seconds = np.arange(48.0, 58.0, 1.0)
    jd = JulianDate(utc=(1973, 12, 29, 23, 59, seconds))
    assert seconds.shape == jd.shape
    for i, second in enumerate(seconds):
        assert jd.tai[i] == JulianDate(utc=(1973, 12, 29, 23, 59, second)).tai

def test_building_JulianDate_from_naive_datetime_raises_exception():
    with assert_raises(ValueError) as info:
        JulianDate(utc=datetime(1973, 12, 29, 23, 59, 48))
    assert 'import timezone' in str(info.exception)

def test_building_JulianDate_from_single_utc_datetime():
    jd = JulianDate(utc=datetime(1973, 12, 29, 23, 59, 48, tzinfo=utc))
    assert jd.tai == 2442046.5

def test_building_JulianDate_from_list_of_utc_datetimes():
    jd = JulianDate(utc=[
        datetime(1973, 12, 29, 23, 59, 48, tzinfo=utc),
        datetime(1973, 12, 30, 23, 59, 48, tzinfo=utc),
        datetime(1973, 12, 31, 23, 59, 48, tzinfo=utc),
        datetime(1974, 1, 1, 23, 59, 47, tzinfo=utc),
        datetime(1974, 1, 2, 23, 59, 47, tzinfo=utc),
        datetime(1974, 1, 3, 23, 59, 47, tzinfo=utc),
        ])
    assert (jd.tai == [
        2442046.5, 2442047.5, 2442048.5, 2442049.5, 2442050.5, 2442051.5,
        ]).all()

def test_indexing_julian_date():
    jd = JulianDate(utc=(1974, 10, range(1, 6)))
    assert jd.shape == (5,)
    jd0 = jd[0]
    assert jd.tai[0] == jd0.tai
    assert jd.tt[0] == jd0.tt
    assert jd.tdb[0] == jd0.tdb
    assert jd.ut1[0] == jd0.ut1
    assert jd.delta_t == jd0.delta_t

def test_slicing_julian_date():
    jd = JulianDate(utc=(1974, 10, range(1, 6)))
    assert jd.shape == (5,)
    jd24 = jd[2:4]
    assert jd24.shape == (2,)
    assert (jd.tai[2:4] == jd24.tai).all()
    assert (jd.tt[2:4] == jd24.tt).all()
    assert (jd.tdb[2:4] == jd24.tdb).all()
    assert (jd.ut1[2:4] == jd24.ut1).all()
    assert jd.delta_t == jd24.delta_t

def test_early_utc():
    jd = JulianDate(utc=(1915, 12, 2, 3, 4, 5.6786786))
    assert abs(jd.tt - 2420833.6283317441) < epsilon
    assert jd.utc_iso() == '1915-12-02T03:04:06Z'

def test_astimezone():
    jd = JulianDate(utc=(1969, 7, 20, 20, 18))
    tz = timezone('US/Eastern')
    dt = jd.astimezone(tz)
    assert dt == tz.localize(datetime(1969, 7, 20, 16, 18, 0, 0))

def test_astimezone_and_leap_second():
    jd = JulianDate(utc=(1969, 7, 20, 20, 18))
    tz = timezone('US/Eastern')
    dt, leap_second = jd.astimezone_and_leap_second(tz)
    assert dt == tz.localize(datetime(1969, 7, 20, 16, 18, 0, 0))
    assert leap_second == 0

def test_utc_datetime():
    jd = JulianDate(utc=(1969, 7, 20, 20, 18))
    dt = jd.utc_datetime()
    assert dt == datetime(1969, 7, 20, 20, 18, 0, 0, utc)

def test_utc_datetime_and_leap_second():
    jd = JulianDate(utc=(1969, 7, 20, 20, 18))
    dt, leap_second = jd.utc_datetime_and_leap_second()
    assert dt == datetime(1969, 7, 20, 20, 18, 0, 0, utc)
    assert leap_second == 0

def test_iso_of_decimal_that_rounds_up():
    jd = JulianDate(utc=(1915, 12, 2, 3, 4, 5.6786786))
    assert jd.utc_iso(places=0) == '1915-12-02T03:04:06Z'
    assert jd.utc_iso(places=1) == '1915-12-02T03:04:05.7Z'
    assert jd.utc_iso(places=2) == '1915-12-02T03:04:05.68Z'
    assert jd.utc_iso(places=3) == '1915-12-02T03:04:05.679Z'
    assert jd.utc_iso(places=4) == '1915-12-02T03:04:05.6787Z'

def test_iso_of_decimal_that_rounds_down():
    jd = JulianDate(utc=(2014, 12, 21, 6, 3, 1.234234))
    assert jd.utc_iso(places=0) == '2014-12-21T06:03:01Z'
    assert jd.utc_iso(places=1) == '2014-12-21T06:03:01.2Z'
    assert jd.utc_iso(places=2) == '2014-12-21T06:03:01.23Z'
    assert jd.utc_iso(places=3) == '2014-12-21T06:03:01.234Z'
    assert jd.utc_iso(places=4) == '2014-12-21T06:03:01.2342Z'

def test_iso_of_leap_second_with_fraction():
    jd = JulianDate(utc=(1973, 12, 31, 23, 59, 60.12349))
    assert jd.utc_iso(places=0) == '1973-12-31T23:59:60Z'
    assert jd.utc_iso(places=1) == '1973-12-31T23:59:60.1Z'
    assert jd.utc_iso(places=2) == '1973-12-31T23:59:60.12Z'
    assert jd.utc_iso(places=3) == '1973-12-31T23:59:60.123Z'
    assert jd.utc_iso(places=4) == '1973-12-31T23:59:60.1235Z'

def test_iso_of_array_showing_whole_seconds():
    jd = JulianDate(utc=(1973, 12, 31, 23, 59, np.arange(58.75, 63.1, 0.5)))
    assert jd.utc_iso(places=0) == [
        '1973-12-31T23:59:59Z',
        '1973-12-31T23:59:59Z',
        '1973-12-31T23:59:60Z',
        '1973-12-31T23:59:60Z',
        '1974-01-01T00:00:00Z',
        '1974-01-01T00:00:00Z',
        '1974-01-01T00:00:01Z',
        '1974-01-01T00:00:01Z',
        '1974-01-01T00:00:02Z',
        ]

def test_iso_of_array_showing_fractions():
    jd = JulianDate(utc=(1973, 12, 31, 23, 59, np.arange(58.75, 63.1, 0.5)))
    assert jd.utc_iso(places=2) == [
        '1973-12-31T23:59:58.75Z',
        '1973-12-31T23:59:59.25Z',
        '1973-12-31T23:59:59.75Z',
        '1973-12-31T23:59:60.25Z',
        '1973-12-31T23:59:60.75Z',
        '1974-01-01T00:00:00.25Z',
        '1974-01-01T00:00:00.75Z',
        '1974-01-01T00:00:01.25Z',
        '1974-01-01T00:00:01.75Z',
        ]

def test_jpl_format():
    jd = JulianDate(utc=(range(-300, 301, 100), 7, 1))
    assert jd.utc_jpl() == [
        'B.C. 0301-Jul-01 00:00:00.0000 UT',
        'B.C. 0201-Jul-01 00:00:00.0000 UT',
        'B.C. 0101-Jul-01 00:00:00.0000 UT',
        'B.C. 0001-Jul-01 00:00:00.0000 UT',
        'A.D. 0100-Jul-01 00:00:00.0000 UT',
        'A.D. 0200-Jul-01 00:00:00.0000 UT',
        'A.D. 0300-Jul-01 00:00:00.0000 UT',
        ]

def test_stftime_of_single_date():
    jd = JulianDate(utc=(1973, 12, 31, 23, 59, 60))
    assert jd.utc_strftime('%Y %m %d %H %M %S') == '1973 12 31 23 59 60'

def test_stftime_of_date_array():
    jd = JulianDate(utc=(1973, 12, 31, 23, 59, np.arange(59.0, 61.1, 1.0)))
    assert jd.utc_strftime('%Y %m %d %H %M %S') == [
        '1973 12 31 23 59 59',
        '1973 12 31 23 59 60',
        '1974 01 01 00 00 00',
        ]

def test_leap_second():

    # During 1973 the offset between UTC and TAI was 12.0 seconds, so
    # TAI should reach the first moment of 1974 while the UTC clock is
    # still reading 12s before midnight (60 - 12 = 48).  Happily, the
    # fraction 0.5 can be precisely represented in floating point, so we
    # can use a bare `==` in this assert:

    t0 = JulianDate(utc=(1973, 12, 31, 23, 59, 48.0)).tai
    assert t0 == 2442048.5

    # Here are some more interesting values:

    t1 = JulianDate(utc=(1973, 12, 31, 23, 59, 58.0)).tai
    t2 = JulianDate(utc=(1973, 12, 31, 23, 59, 59.0)).tai
    t3 = JulianDate(utc=(1973, 12, 31, 23, 59, 60.0)).tai
    t4 = JulianDate(utc=(1974, 1, 1, 0, 0, 0.0)).tai
    t5 = JulianDate(utc=(1974, 1, 1, 0, 0, 1.0)).tai

    # The step from 23:59:59 to 0:00:00 is here a two-second step,
    # because of the leap second 23:59:60 that falls in between:

    assert abs(t4 - t2 - 2.0 * one_second) < epsilon

    # Thus, the five dates given above are all one second apart:

    assert abs(t2 - t1 - one_second) < epsilon
    assert abs(t3 - t2 - one_second) < epsilon
    assert abs(t4 - t3 - one_second) < epsilon
    assert abs(t5 - t4 - one_second) < epsilon

    # And all these dates can be converted back to UTC.

    assert JulianDate(tai=t0).utc_iso() == '1973-12-31T23:59:48Z'
    assert JulianDate(tai=t1).utc_iso() == '1973-12-31T23:59:58Z'
    assert JulianDate(tai=t2).utc_iso() == '1973-12-31T23:59:59Z'
    assert JulianDate(tai=t3).utc_iso() == '1973-12-31T23:59:60Z'
    assert JulianDate(tai=t4).utc_iso() == '1974-01-01T00:00:00Z'
    assert JulianDate(tai=t5).utc_iso() == '1974-01-01T00:00:01Z'



"""Compare the output of Skyfield with the routines from NOVAS for keplerian orbiting bodies"""

import skyfield.keplerianlib
from skyfield.keplerianlib import KeplerianOrbit, ICRCoordinates

from ..timelib import JulianDate, julian_date

DISTANCE_EPSILON = 0.026

def test_semimajorAxisToOrbitalPeriod():
    assert skyfield.keplerianlib.semimajorAxisToOrbitalPeriod(1) == 1
    assert skyfield.keplerianlib.semimajorAxisToOrbitalPeriod(1.523679) == 1.8807896358663763
    assert skyfield.keplerianlib.semimajorAxisToOrbitalPeriod(4.27371348392) == 8.835031547398543

def test_orbitalPeriodToSemimajorAxis():
    assert skyfield.keplerianlib.orbitalPeriodToSemimajorAxis(1) == 1
    assert skyfield.keplerianlib.orbitalPeriodToSemimajorAxis(1.8807896358663763) == 1.523679
    assert skyfield.keplerianlib.orbitalPeriodToSemimajorAxis(8.835031547398543) == 4.27371348392

def test_convergeEccentricAnomaly():
    test = skyfield.keplerianlib.convergeEccentricAnomaly(
        hoyle_8077['mean_anomaly'],
        hoyle_8077['eccentricity'],
        15
    )
    assert test == hoyle_8077['eccentric_anomaly']

def test_instantiate_8077_hoyle():
    hoyle = KeplerianOrbit( hoyle_8077['semimajor_axis'],
                            hoyle_8077['eccentricity'],
                            hoyle_8077['inclination'],
                            hoyle_8077['longitude_ascending'],
                            hoyle_8077['argument_perihelion'],
                            hoyle_8077['mean_anomaly'],
                            hoyle_8077['epoch'])

    assert hoyle != None

def test_instantiate_coordinates():
    coords = ICRCoordinates(x=500.25, y=10.76, z=0.1125)

    assert coords != None

def test_coordinatesEquivalence():
    coords_the_first = ICRCoordinates(x=500.25, y=10.76, z=0.1125)
    coords_the_second = ICRCoordinates(x=500.25, y=10.76, z=0.1125)

    assert coords_the_first.equalTo(coords_the_second)

"""
 data gotten from Horizons
date: 2456517.500000000 = A.D. 2013-Aug-13 00:00:00.0000 (CT)
expected coords (AU) 2.421251132790093E+00 -1.918893156489506E+00 -9.813409585464707E-02

 Horizon Params
Ephemeris Type [change] :   VECTORS
Target Body [change] :  Asteroid 8077 Hoyle (1986 AW2)
Coordinate Origin [change] :    Solar System Barycenter (SSB) [500@0]
Time Span [change] :    Start=2013-08-13, Stop=2013-09-12, Step=1 d
Table Settings [change] :   defaults
Display/Output [change] :   default (formatted HTML)

EPOCH=  2453995.5 ! 2006-Sep-17.00 (CT)          Residual RMS= .43359
   EC= .2110946491840378   QR= 2.077692130214496   TP= 2454360.1855338747
   OM= 135.855972529608    W=  34.4378477722205    IN= 17.25814783060462
   A= 2.633639292806857    MA= 275.9015153135912   ADIST= 3.189586455399217
   PER= 4.27408            N= .230605479           ANGMOM= .027287332
   DAN= 2.14316            DDN= 3.04671            L= 169.07331
   B= 9.658454799999999    MOID= 1.10581994        TP= 2007-Sep-16.6855338747

"""

hoyle_8077 = {
    'semimajor_axis' : 2.633278254269645,
    'eccentricity' : .2109947010748546,
    'inclination' : 17.25945395594321,
    'longitude_ascending' : 135.8512354853258,
    'argument_perihelion' : 34.46503170092878,
    'mean_anomaly' : 330.9918926661418,
    'eccentric_anomaly' : 4.0942988262501965,
    'epoch' : JulianDate(tt=(2007, 5, 14)),
}


def test_get_8077_hoyle_ecliptic_on_dev_sprint_day_2():
    hoyle = KeplerianOrbit( hoyle_8077['semimajor_axis'],
                            hoyle_8077['eccentricity'],
                            hoyle_8077['inclination'],
                            hoyle_8077['longitude_ascending'],
                            hoyle_8077['argument_perihelion'],
                            hoyle_8077['mean_anomaly'],
                            hoyle_8077['epoch'])

    date = JulianDate(tt=(2013, 8, 13))
    # print date.tt

    test = hoyle.getECLCoordinatesOnJulianDate(date)

    #print test
    epsilon = 2e-2
    assert abs(test.x - 2.421251271197979) < epsilon
    assert abs(test.y - -1.918893007049262) < epsilon
    assert abs(test.z - -0.09813403009731327) < epsilon



"""Python classes that represent various classes of star."""

from numpy import array, cos, outer, sin
from .constants import AU_KM, ASEC2RAD, C, C_AUDAY, DAY_S, T0
from .functions import length_of
from .positionlib import Astrometric
from .relativity import light_time_difference
from .units import Angle

class Star(object):

    au_km = AU_KM

    def __init__(self, ra=None, dec=None, ra_hours=None, dec_degrees=None,
                 ra_mas_per_year=0.0, dec_mas_per_year=0.0,
                 parallax_mas=0.0, radial_km_per_s=0.0, names=()):

        if ra_hours is not None:
            self.ra = Angle(hours=ra_hours)
        elif isinstance(ra, Angle):
            self.ra = ra
        else:
            raise TypeError('please provide either ra_hours=<float> or else'
                            ' ra=<skyfield.units.Angle object>')

        if dec_degrees is not None:
            self.dec = Angle(degrees=dec_degrees)
        elif isinstance(dec, Angle):
            self.dec = dec
        else:
            raise TypeError('please provide either dec_degrees=<float> or else'
                            ' dec=<skyfield.units.Angle object>')

        self.ra_mas_per_year = ra_mas_per_year
        self.dec_mas_per_year = dec_mas_per_year
        self.parallax_mas = parallax_mas
        self.radial_km_per_s = radial_km_per_s
        self.names = names

        self._compute_vectors()

    def __repr__(self):
        opts = []
        for name in ['ra_mas_per_year', 'dec_mas_per_year',
                     'parallax_mas', 'radial_km_per_s', 'names']:
            value = getattr(self, name)
            if value:
                opts.append(', {0}={1!r}'.format(name, value))
        return 'Star(ra_hours={0!r}, dec_degrees={1!r}{2})'.format(
            self.ra.hours, self.dec.degrees, ''.join(opts))

    def _observe_from_bcrs(self, observer):
        position, velocity = self._position_au, self._velocity_au_per_d
        jd = observer.jd
        dt = light_time_difference(position, observer.position.au)
        if jd.shape:
            position = (outer(velocity, jd.tdb + dt - T0).T + position).T
        else:
            position = position + velocity * (jd.tdb + dt - T0)
        vector = position - observer.position.au
        distance = length_of(vector)
        light_time = distance / C_AUDAY

        g = Astrometric(vector, (observer.velocity.au_per_d.T - velocity).T, jd)
        g.observer = observer
        g.distance = distance
        g.light_time = light_time
        return g

    def _compute_vectors(self):
        """Compute the star's position as an ICRS position and velocity."""

        # Use 1 gigaparsec for stars whose parallax is zero.

        parallax = self.parallax_mas
        if parallax <= 0.0:
            parallax = 1.0e-6

        # Convert right ascension, declination, and parallax to position
        # vector in equatorial system with units of au.

        dist = 1.0 / sin(parallax * 1.0e-3 * ASEC2RAD)
        r = self.ra.radians
        d = self.dec.radians
        cra = cos(r)
        sra = sin(r)
        cdc = cos(d)
        sdc = sin(d)

        self._position_au = array((
            dist * cdc * cra,
            dist * cdc * sra,
            dist * sdc,
            ))

        # Compute Doppler factor, which accounts for change in light
        # travel time to star.

        k = 1.0 / (1.0 - self.radial_km_per_s / C * 1000.0)

        # Convert proper motion and radial velocity to orthogonal
        # components of motion with units of au/day.

        pmr = self.ra_mas_per_year / (parallax * 365.25) * k
        pmd = self.dec_mas_per_year / (parallax * 365.25) * k
        rvl = self.radial_km_per_s * DAY_S / self.au_km * k

        # Transform motion vector to equatorial system.

        self._velocity_au_per_d = array((
            - pmr * sra - pmd * sdc * cra + rvl * cdc * cra,
              pmr * cra - pmd * sdc * sra + rvl * cdc * sra,
              pmd * cdc + rvl * sdc,
              ))



from skyfield.units import Angle

def test_degree_rounding():
    tenth = 0.1 / 60.0 / 60.0  # of an arcsecond

    assert str(Angle(degrees=tenth * -600.75)) == '-00deg 01\' 00.1"'
    assert str(Angle(degrees=tenth * -600.25)) == '-00deg 01\' 00.0"'
    assert str(Angle(degrees=tenth * -599.75)) == '-00deg 01\' 00.0"'
    assert str(Angle(degrees=tenth * -599.25)) == '-00deg 00\' 59.9"'

    assert str(Angle(degrees=tenth * -1.75)) == '-00deg 00\' 00.2"'
    assert str(Angle(degrees=tenth * -1.25)) == '-00deg 00\' 00.1"'
    assert str(Angle(degrees=tenth * -0.75)) == '-00deg 00\' 00.1"'
    assert str(Angle(degrees=tenth * -0.25)) == '-00deg 00\' 00.0"'

    assert str(Angle(degrees=0.0)) == '00deg 00\' 00.0"'

    assert str(Angle(degrees=tenth * 0.25)) == '00deg 00\' 00.0"'
    assert str(Angle(degrees=tenth * 0.75)) == '00deg 00\' 00.1"'
    assert str(Angle(degrees=tenth * 1.25)) == '00deg 00\' 00.1"'
    assert str(Angle(degrees=tenth * 1.75)) == '00deg 00\' 00.2"'

    assert str(Angle(degrees=tenth * 599.25)) == '00deg 00\' 59.9"'
    assert str(Angle(degrees=tenth * 599.75)) == '00deg 01\' 00.0"'
    assert str(Angle(degrees=tenth * 600.25)) == '00deg 01\' 00.0"'
    assert str(Angle(degrees=tenth * 600.75)) == '00deg 01\' 00.1"'



"""Top-level objects and functions offered by the Skyfield library.

Importing this ``skyfield.api`` module causes Skyfield to load up the
default JPL planetary ephemeris ``de421`` and create planet objects like
``earth`` and ``mars`` that are ready for your use.

"""
from datetime import datetime
from math import pi
from .constants import tau
from .errors import DeprecationError
from .iokit import load
from .starlib import Star
from .timelib import JulianDate, T0, now, utc
from .toposlib import Topos
from .units import Angle
from .named_stars import NamedStar

__all__ = ['Angle', 'JulianDate', 'NamedStar', 'Star', 'Topos',
           'datetime', 'load', 'now', 'utc', 'T0', 'pi', 'tau']

# An attempt at friendliest-possible deprecations:

class DeprecatedPlanet(object):
    def __init__(self, name):
        self.name = name

    def __call__(self, jd):
        raise DeprecationError("""Skyfield no longer auto-loads the planets.

If you simply want your old Skyfield script to start working again,
downgrade to Skyfield version 0.4 using a command like:

        pip install skyfield==0.4

To upgrade your script to a modern version of Skyfield, find each place
that you loaded a body from the skyfield.api module and called it:

        from skyfield.api import {body}
        position = {body}(jd)

Instead, load the ephemeris like DE421 explicitly, look up the body in
the ephemeris, and use the method at() to generate a position:

        from skyfield.api import load
        planets = load('de421.bsp')
        {body} = planets['{body}']
        position = {body}.at(jd)

More documentation can be found at: http://rhodesmill.org/skyfield/"""
                               .format(body=self.name))

class DeprecatedEarth(DeprecatedPlanet):
    def topos(self, *args, **kw):
        raise DeprecationError("""Skyfield no longer auto-loads an ephemeris.

If you simply want your old Skyfield script to start working again,
downgrade to Skyfield version 0.4 using a command like:

        pip install skyfield==0.4

To upgrade your script to a modern version of Skyfield, find each place
you loaded the Earth from skyfield.api and called its topos() method:

        from skyfield.api import earth
        place = earth.topos(...)

Instead, load the ephemeris like DE421 explicitly, look up Earth in the
ephemeris, and use the method at() to generate a position:

        from skyfield.api import load
        planets = load('de421.bsp')
        earth = planets['earth']
        place = earth.topos(...)

More documentation can be found at: http://rhodesmill.org/skyfield/""")

sun = DeprecatedPlanet('sun')
moon = DeprecatedPlanet('moon')
mercury = DeprecatedPlanet('mercury')
venus = DeprecatedPlanet('venus')
earth = DeprecatedEarth('earth')
mars = DeprecatedPlanet('mars')
jupiter = DeprecatedPlanet('jupiter')
saturn = DeprecatedPlanet('saturn')
uranus = DeprecatedPlanet('uranus')
neptune = DeprecatedPlanet('neptune')
pluto = DeprecatedPlanet('pluto')



from numpy import arcsin, arctan2, array, cos, sin, sqrt
from skyfield.constants import tau

def dots(v, u):
    """Given one or more vectors in `v` and `u`, return their dot products.

    This works whether `v` and `u` each have the shape ``(3,)``, or
    whether they are each whole arrays of corresponding x, y, and z
    coordinates and have shape ``(3, N)``.

    """
    return (v * u).sum(axis=0)

def length_of(xyz):
    """Given a 3-element array `[x y z]`, return its length.

    The three elements can be simple scalars, or the array can be two
    dimensions and offer three whole series of x, y, and z coordinates.

    """
    return sqrt((xyz * xyz).sum(axis=0))

def to_polar(xyz):
    """Convert ``[x y z]`` into spherical coordinates ``(r, theta, phi)``.

    ``r`` - vector length
    ``theta`` - angle above (+) or below (-) the xy-plane
    ``phi`` - angle around the z-axis

    The order of the three return values is intended to match ISO 31-11.

    """
    r = length_of(xyz)
    x, y, z = xyz
    theta = arcsin(z / r)
    phi = arctan2(y, x) % tau
    return r, theta, phi

def from_polar(r, theta, phi):
    """Convert ``(r, theta, phi)`` to Cartesian coordinates ``[x y z]``.

    ``r`` - vector length
    ``theta`` - angle above (+) or below (-) the xy-plane
    ``phi`` - angle around the z-axis

    The order of the three arguments is intended to match ISO 31-11.

    """
    rxy = r * cos(theta)
    return array((rxy * cos(phi), rxy * sin(phi), r * sin(theta)))

def rot_x(theta):
    c = cos(theta)
    s = sin(theta)
    return array([(1.0, 0.0, 0.0), (0.0, c, -s), (0.0, s, c)])

def rot_y(theta):
    c = cos(theta)
    s = sin(theta)
    return array([(c, 0.0, s), (0.0, 1.0, 0.0), (-s, 0.0, c)])

def rot_z(theta):
    c = cos(theta)
    s = sin(theta)
    zero = theta * 0.0
    one = zero + 1.0
    return array(((c, -s, zero), (s, c, zero), (zero, zero, one)))



"""Tests of whether units behave."""

from assay import assert_raises
from numpy import array
from skyfield import units

try:
    from astropy import units as u
except ImportError:
    u = None

# needs_astropy = pytest.mark.skipif(u is None, reason='cannot import AstroPy')

def test_stringifying_vector_distance():
    a = array([1.23, 4.56])
    assert str(units.Distance(au=a)) == '[ 1.23  4.56] au'

def test_iterating_over_raw_measurement():
    distance = units.Distance(au=1.234)
    with assert_raises(units.UnpackingError):
        x, y, z = distance

def test_iterating_over_raw_velocity():
    velocity = units.Velocity(au_per_d=1.234)
    with assert_raises(units.UnpackingError):
        x, y, z = velocity

def test_converting_from_km_to_m():
    distance = units.Distance(km=1.234)
    assert abs(distance.m - 1234.0) < 1e-15

def test_converting_from_m_to_km():
    distance = units.Distance(m=1234.0)
    assert abs(distance.km - 1.234) < 1e-15

# @needs_astropy
def test_converting_distance_with_astropy():
    distance = units.Distance(au=1.234)
    value1 = distance.km
    value2 = distance.to(u.km)
    epsilon = 0.02         # definitions of AU seem to disagree slightly
    assert abs(value1 - value2.value) < epsilon

# @needs_astropy
def test_converting_velocity_with_astropy():
    velocity = units.Velocity(au_per_d=1.234)
    value1 = velocity.km_per_s
    value2 = velocity.to(u.km / u.s)
    epsilon = 1e-6
    assert abs(value1 - value2.value) < epsilon
